# -*- coding: utf-8 -*-

# spechomo, Spectral homogenization of multispectral satellite data
#
# Copyright (C) 2020  Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
#
# This software was developed within the context of the GeoMultiSens project funded
# by the German Federal Ministry of Education and Research
# (project grant code: 01 IS 14 010 A-C).
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version. Please note the following exception: `spechomo` depends on tqdm,
# which is distributed under the Mozilla Public Licence (MPL) v2.0 except for the
# files "tqdm/_tqdm.py", "setup.py", "README.rst", "MANIFEST.in" and ".gitignore".
# Details can be found here: https://github.com/tqdm/tqdm/blob/master/LICENCE.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Main module."""

import os
import numpy as np
import logging  # noqa F401  # flake8 issue
from typing import Union, Tuple  # noqa F401  # flake8 issue
from multiprocessing import cpu_count
import traceback
import time
from geoarray import GeoArray  # noqa F401  # flake8 issue
from specclassify import classify_image
# from specclassify import kNN_MinimumDistance_Classifier

from .classifier import Cluster_Learner
from .exceptions import ClassifierNotAvailableError
from .logging import SpecHomo_Logger
from .options import options


__author__ = 'Daniel Scheffler'

_classifier_rootdir = options['classifiers']['rootdir']


class SpectralHomogenizer(object):
    """Class for applying spectral homogenization by applying an interpolation or machine learning approach."""

    def __init__(self, classifier_rootDir='', logger=None, CPUs=None):
        """Get instance of SpectralHomogenizer.

        :param classifier_rootDir:  root directory where machine learning classifiers are stored.
        :param logger:              instance of logging.Logger
        """
        self.classifier_rootDir = classifier_rootDir or _classifier_rootdir
        self.logger = logger or SpecHomo_Logger(__name__)
        self.CPUs = CPUs or cpu_count()

    def interpolate_cube(self, arrcube, source_CWLs, target_CWLs, kind='linear'):
        # type: (Union[np.ndarray, GeoArray], list, list, str) -> GeoArray
        """Spectrally interpolate the spectral bands of a remote sensing image to new band positions.

        :param arrcube:     array to be spectrally interpolated
        :param source_CWLs: list of source central wavelength positions
        :param target_CWLs: list of target central wavelength positions
        :param kind:        interpolation kind to be passed to scipy.interpolate.interp1d (default: 'linear')
        :return:
        """
        from scipy.interpolate import interp1d  # import here to avoid static TLS ImportError

        assert kind in ['linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic'], \
            "%s is not a supported kind of spectral interpolation." % kind
        assert arrcube is not None,\
            'L2B_obj.interpolate_cube_linear expects a numpy array as input. Got %s.' % type(arrcube)

        orig_CWLs, target_CWLs = np.array(source_CWLs), np.array(target_CWLs)

        self.logger.info(
            'Performing spectral homogenization (%s interpolation) with target wavelength positions at %s nm.'
            % (kind, ', '.join(np.round(np.array(target_CWLs[:-1]), 1).astype(str)) +
               ' and %s' % np.round(target_CWLs[-1], 1)))
        outarr = interp1d(np.array(orig_CWLs), arrcube, axis=2, kind=kind, fill_value='extrapolate')(target_CWLs)

        if np.min(outarr) >= np.iinfo(np.int16).min and np.max(outarr) <= np.iinfo(np.int16).max:
            outarr = outarr.astype(np.int16)
        elif np.min(outarr) >= np.iinfo(np.int32).min and np.max(outarr) <= np.iinfo(np.int32).max:
            outarr = outarr.astype(np.int32)
        else:
            raise TypeError('The interpolated data cube cannot be cast into a 16- or 32-bit integer array.')

        assert outarr.shape == tuple([*arrcube.shape[:2], len(target_CWLs)])

        return GeoArray(outarr)

    def predict_by_machine_learner(self, arrcube, method, src_satellite, src_sensor, src_LBA, tgt_satellite, tgt_sensor,
                                   tgt_LBA, n_clusters=50, classif_alg='MinDist', kNN_n_neighbors=10,
                                   global_clf_threshold=options['classifiers']['prediction']['global_clf_threshold'],
                                   src_nodataVal=None, out_nodataVal=None, compute_errors=False, bandwise_errors=True,
                                   fallback_argskwargs=None):
        # type: (Union[np.ndarray, GeoArray], str, str, str, list, str, str, list, int, str, int, Union[str, int, float], int, int, bool, bool, dict) -> tuple  # noqa
        """Predict spectral bands of target sensor by applying a machine learning approach.

        NOTE:   You may use the function spechomo.utils.list_available_transformations() to get a list of available
                transformations. You may also copy the input parameters for this method from the output there.

        :param arrcube:             input image array for target sensor spectral band prediction (rows x cols x bands)
        :param method:              machine learning approach to be used for spectral bands prediction
                                    'LR': Linear Regression
                                    'RR': Ridge Regression
                                    'QR': Quadratic Regression
                                    'RFR':  Random Forest Regression  (50 trees; does not allow spectral sub-clustering)
        :param src_satellite:       source satellite, e.g., 'Landsat-8'
        :param src_sensor:          source sensor, e.g., 'OLI_TIRS'
        :param src_LBA:             source LayerBandsAssignment  # TODO document this
        :param tgt_satellite:       target satellite, e.g., 'Landsat-8'
        :param tgt_sensor:          target sensor, e.g., 'OLI_TIRS'
        :param tgt_LBA:             target LayerBandsAssignment  # TODO document this
        :param n_clusters:          Number of spectral clusters to be used during LR/ RR/ QR homogenization.
                                    E.g., 50 means that the image to be converted to the spectral target sensor
                                    is clustered into 50 spectral clusters and one separate machine learner per
                                    cluster is applied to the input data to predict the homogenized image. If
                                    'spechomo_n_clusters' is set to 1, the source image is not clustered and
                                    only one machine learning classifier is used for prediction.
        :param classif_alg:         Multispectral classification algorithm to be used to determine the spectral cluster
                                    each pixel belongs to.
                                    'MinDist': Minimum Distance (Nearest Centroid)
                                    'kNN': k-nearest-neighbour
                                    'kNN_MinDist': k-nearest-neighbour Minimum Distance (Nearest Centroid)
                                    'SAM': spectral angle mapping
                                    'kNN_SAM': k-nearest-neighbour spectral angle mapping
                                    'SID': spectral information divergence
                                    'FEDSA': fused euclidian distance / spectral angle
                                    'kNN_FEDSA': k-nearest-neighbour fused euclidian distance / spectral angle
        :param kNN_n_neighbors:     The number of neighbors to be considered in case 'classif_alg' is set to 'kNN'.
                                    Otherwise, this parameter is ignored.
        :param global_clf_threshold:  If given, all pixels where the computed similarity metric (set by 'classif_alg')
                                      exceeds the given threshold are predicted using the global classifier (based on a
                                      single transformation per band).
                                      - only usable for 'MinDist', 'SAM' and 'SID'
                                      - may be given as float, integer or string to label a certain distance percentile
                                      - if given as string, it must match the format, e.g., '10%' for labelling the
                                      worst 10 % of the distances as unclassified
        :param src_nodataVal:       no data value of source image (arrcube)
                                    - if no nodata value is set, it is tried to be auto-computed from arrcube
        :param out_nodataVal:       no data value of predicted image
        :param compute_errors:      whether to compute pixel- / bandwise model errors for estimated pixel values
                                    (default: false)
        :param bandwise_errors      whether to compute error information for each band separately (True - default)
                                    or to average errors over bands using median (False) (ignored in case of fallback)
        :param fallback_argskwargs: arguments and keyword arguments to be passed to the fallback algorithm
                                    SpectralHomogenizer.interpolate_cube() in case harmonization fails
        :return:                    predicted array (rows x columns x bands)
        :rtype:                     Tuple[np.ndarray, Union[np.ndarray, None]]
        """
        # TODO: add LBA validation to .predict()
        kw = dict(method=method,
                  classifier_rootDir=self.classifier_rootDir,
                  n_clusters=n_clusters,
                  classif_alg=classif_alg,
                  CPUs=self.CPUs,
                  logger=self.logger)

        if classif_alg.startswith('kNN'):
            kw['n_neighbors'] = kNN_n_neighbors

        RSI_CP = RSImage_ClusterPredictor(**kw)

        ######################
        # get the classifier #
        ######################

        cls = None
        exc = Exception()
        try:
            cls = RSI_CP.get_classifier(src_satellite, src_sensor, src_LBA, tgt_satellite, tgt_sensor, tgt_LBA)

        except FileNotFoundError as e:
            self.logger.warning('No machine learning classifier available that fulfills the specifications of the '
                                'spectral reference sensor. Falling back to linear interpolation for performing '
                                'spectral homogenization.')
            exc = e

        except ClassifierNotAvailableError as e:
            self.logger.error('\nAn error occurred during spectral homogenization using the %s classifier. '
                              'Falling back to linear interpolation. Error message was: ' % method)
            self.logger.error(traceback.format_exc())
            exc = e

        ##################
        # run prediction #
        ##################

        errors = None
        if cls:
            self.logger.info('Performing spectral homogenization using %s. Target is %s %s %s.'
                             % (method, tgt_satellite, tgt_sensor, tgt_LBA))
            im_homo = RSI_CP.predict(arrcube,
                                     classifier=cls,
                                     in_nodataVal=src_nodataVal,
                                     cmap_nodataVal=src_nodataVal,
                                     out_nodataVal=out_nodataVal,
                                     global_clf_threshold=global_clf_threshold)  # type: GeoArray

            if compute_errors:
                errors = RSI_CP.compute_prediction_errors(im_homo, cls,
                                                          nodataVal=src_nodataVal,
                                                          cmap_nodataVal=src_nodataVal)

                if not bandwise_errors:
                    errors = np.median(errors, axis=2).astype(errors.dtype)

        elif fallback_argskwargs:
            # fallback: use linear interpolation and set errors to an array of zeros
            im_homo = self.interpolate_cube(**fallback_argskwargs)  # type: GeoArray

            if compute_errors:
                self.logger.warning("Spectral homogenization algorithm had to be performed by linear interpolation "
                                    "(fallback). Unable to compute any accuracy information from that.")
                if bandwise_errors:
                    errors = np.zeros_like(im_homo, dtype=np.int16)
                else:
                    errors = np.zeros(im_homo.shape[:2], dtype=np.int16)

        else:
            raise exc

        # add metadata
        im_homo.metadata.band_meta['wavelength'] = cls.tgt_wavelengths if cls else fallback_argskwargs['target_CWLs']
        im_homo.classif_map = RSI_CP.classif_map
        im_homo.distance_metrics = RSI_CP.distance_metrics

        # handle negative values in the predicted image => set these pixels to nodata
        # im_homo = set_negVals_to_nodata(im_homo, out_nodataVal)

        return im_homo, errors


class RSImage_ClusterPredictor(object):
    """Predictor class applying the predict() function of a machine learning classifier described by the given args."""

    def __init__(self, method='LR', n_clusters=50, classif_alg='MinDist', classifier_rootDir='',
                 CPUs=1, logger=None, **kw_clf_init):
        # type: (str, int, str, str, Union[None, int], logging.Logger, dict) -> None
        """Get an instance of RSImage_ClusterPredictor.

        :param method:              machine learning approach to be used for spectral bands prediction
                                    'LR':   Linear Regression
                                    'RR':   Ridge Regression
                                    'QR':   Quadratic Regression
                                    'RFR':  Random Forest Regression  (50 trees; does not allow spectral sub-clustering)
        :param n_clusters:          Number of spectral clusters to be used during LR/ RR/ QR homogenization.
                                    E.g., 50 means that the image to be converted to the spectral target sensor
                                    is clustered into 50 spectral clusters and one separate machine learner per
                                    cluster is applied to the input data to predict the homogenized image. If
                                    'n_clusters' is set to 1, the source image is not clustered and
                                    only one machine learning classifier is used for prediction.
        :param classif_alg:         algorithm to be used for image classification
                                    (to define which cluster each pixel belongs to)
                                    'MinDist': Minimum Distance (Nearest Centroid)
                                    'kNN': k-nearest-neighbour
                                    'kNN_MinDist': k-nearest-neighbour Minimum Distance (Nearest Centroid)
                                    'SAM': spectral angle mapping
                                    'kNN_SAM': k-nearest-neighbour spectral angle mapping
                                    'SID': spectral information divergence
                                    'FEDSA': fused euclidian distance / spectral angle
                                    'kNN_FEDSA': k-nearest-neighbour fused euclidian distance / spectral angle
        :param classifier_rootDir:  root directory where machine learning classifiers are stored.
        :param CPUs:                number of CPUs to use (default: 1)
        :param logger:              instance of logging.Logger()
        :param kw_clf_init          keyword arguments to be passed to classifier init functions if possible,
                                    e.g., 'n_neighbours' sets the number of neighbours to be considered in kNN
                                    classification algorithms (set by 'classif_alg')
        """
        self.method = method
        self.n_clusters = n_clusters
        self.classifier_rootDir = os.path.abspath(classifier_rootDir) if classifier_rootDir else _classifier_rootdir
        self.classif_map = None
        self.classif_map_fractions = None
        self.distance_metrics = None
        self.CPUs = CPUs or cpu_count()
        self.classif_alg = classif_alg
        self.logger = logger or SpecHomo_Logger(__name__)  # must be pickable
        self.kw_clf_init = kw_clf_init

        # validate
        if method == 'RFR' and n_clusters > 1:
            self.logger.warning("The spectral homogenization method 'Random Forest Regression' does not allow spectral "
                                "sub-clustering. Setting 'n_clusters' to 1.")
            self.n_clusters = 1

        if self.classif_alg.startswith('kNN') and \
           'n_neighbors' in kw_clf_init and \
           self.n_clusters < kw_clf_init['n_neighbors']:
            self.kw_clf_init['n_neighbors'] = self.n_clusters

    def get_classifier(self, src_satellite, src_sensor, src_LBA, tgt_satellite, tgt_sensor, tgt_LBA):
        # type: (str, str, list, str, str, list) -> Cluster_Learner
        """Select the correct machine learning classifier out of previously saved classifier collections.

        Describe the classifier specifications with the given arguments.
        :param src_satellite:   source satellite, e.g., 'Landsat-8'
        :param src_sensor:      source sensor, e.g., 'OLI_TIRS'
        :param src_LBA:         source LayerBandsAssignment
        :param tgt_satellite:   target satellite, e.g., 'Landsat-8'
        :param tgt_sensor:      target sensor, e.g., 'OLI_TIRS'
        :param tgt_LBA:         target LayerBandsAssignment
        :return:                classifier instance loaded from disk
        """
        args_fd = (self.classifier_rootDir, self.method, self.n_clusters,
                   src_satellite, src_sensor, src_LBA, tgt_satellite, tgt_sensor, tgt_LBA)

        try:
            CL = Cluster_Learner.from_disk(*args_fd)

        except FileNotFoundError:
            if self.classifier_rootDir == _classifier_rootdir:
                # the default root directory is used

                if not os.path.exists(os.path.join(_classifier_rootdir, '%s_classifiers.zip' % self.method)):
                    # download the classifiers
                    self.logger.info('The pre-trained classifiers have not been downloaded yet. Downloading...')

                    from .utils import download_pretrained_classifiers
                    download_pretrained_classifiers(method=self.method,
                                                    tgt_dir=self.classifier_rootDir)

                else:
                    self.logger.error('%s classifiers found at %s. However, they do not contain a suitable classifier '
                                      'for the current predition. If desired, delete the existing classifiers and try '
                                      'again. Pre-trained classifiers are then automatically downloaded.'
                                      % (self.method, self.classifier_rootDir))

                # try again
                CL = Cluster_Learner.from_disk(*args_fd)

            else:
                # classifier not found in the user provided root directory
                raise

        return CL

    def predict(self, image, classifier, in_nodataVal=None, out_nodataVal=None, cmap_nodataVal=None,
                global_clf_threshold=None, unclassified_pixVal=-1):
        # type: (Union[np.ndarray, GeoArray], Cluster_Learner, float, float, float, Union[str, int, float], int) -> GeoArray  # noqa
        """Apply the prediction function of the given specifier to the given remote sensing image.

        # NOTE: The 'nodataVal' is written

        :param image:           3D array representing the input image
        :param classifier:      the classifier instance
        :param in_nodataVal:    no data value of the input image
                                (auto-computed if not given or contained in image GeoArray)
        :param out_nodataVal:   no data value written into the predicted image
                                (copied from the input image if not given)
        :param cmap_nodataVal:  no data value for the classification map
                                in case more than one sub-classes are used for prediction
        :param global_clf_threshold:  If given, all pixels where the computed similarity metric (set by 'classif_alg')
                                      exceeds the given threshold are predicted using the global classifier (based on a
                                      single transformation per band).
                                      - not usable for 'kNN'
                                      - may be given as float, integer or string to label a certain distance percentile
                                      - if given as string, it must match the format, e.g., '10%' for labelling the
                                      worst 10 % of the distances as unclassified
        :param unclassified_pixVal:     pixel value to be used in the classification map for unclassified pixels
                                        (default: -1)
        :return:                3D array representing the predicted spectral image cube
        """
        image = image if isinstance(image, GeoArray) else GeoArray(image, nodata=in_nodataVal)

        # ensure image.nodata is present (important for classify_image() -> overwrites cmap at nodata positions)
        image.nodata = in_nodataVal if in_nodataVal is not None else image.nodata  # might be auto-computed here

        ##########################
        # get classification map #
        ##########################

        # assign each input pixel to a cluster (compute classification with cluster centers as endmembers)
        if self.classif_map is None:
            if self.n_clusters > 1:
                t0 = time.time()
                kw_clf = dict(classif_alg=self.classif_alg,
                              in_nodataVal=image.nodata,
                              cmap_nodataVal=cmap_nodataVal,  # written into classif_map at nodata
                              CPUs=self.CPUs,
                              return_distance=True,
                              **self.kw_clf_init)

                if self.classif_alg in ['MinDist', 'kNN_MinDist', 'SAM', 'kNN_SAM', 'SID', 'FEDSA', 'kNN_FEDSA']:
                    kw_clf.update(dict(unclassified_threshold=global_clf_threshold,
                                       unclassified_pixVal=unclassified_pixVal))

                if self.classif_alg == 'RF':
                    train_spectra = np.vstack([classifier.MLdict[clust].cluster_sample_spectra
                                               for clust in range(classifier.n_clusters)])
                    train_labels = list(np.hstack([[i] * 100 for i in range(classifier.n_clusters)]))
                else:
                    train_spectra = classifier.cluster_centers
                    train_labels = classifier.cluster_pixVals

                self.classif_map, self.distance_metrics = classify_image(image, train_spectra, train_labels, **kw_clf)

                # compute spectral distance
                # dist = kNN_MinimumDistance_Classifier.compute_euclidian_distance_3D(image, train_spectra)
                # idxs = self.classif_map.reshape(-1, self.classif_map.shape[2])
                # self.distance_metrics = \
                #     dist.reshape(-1, dist.shape[2])[np.arange(dist.shape[0] * dist.shape[1])[:, np.newaxis], idxs] \
                #         .reshape(self.classif_map.shape)
                # print('ED MAX MIN:', self.distance_metrics.max(), self.distance_metrics.min())

                self.logger.info('Total classification time: %s'
                                 % time.strftime("%H:%M:%S", time.gmtime(time.time() - t0)))

            else:
                self.classif_map = GeoArray(np.full((image.rows, image.cols), classifier.cluster_pixVals[0], np.int16),
                                            nodata=cmap_nodataVal)

                # overwrite all pixels where the input image contains nodata in ANY band
                # (would lead to faulty predictions due to multivariate prediction algorithms)
                if in_nodataVal is not None and cmap_nodataVal is not None:
                    self.classif_map[np.any(image[:] == image.nodata, axis=2)] = cmap_nodataVal

                self.distance_metrics = np.zeros_like(self.classif_map, np.float32)

        ####################
        # apply prediction #
        ####################

        # adjust classifier
        if self.CPUs is None or self.CPUs > 1:
            # FIXME does not work -> parallelize with https://github.com/ajtulloch/sklearn-compiledtrees?
            classifier.n_jobs = cpu_count() if self.CPUs is None else self.CPUs

        # NOTE: prediction is applied in 1000 x 1000 tiles to save memory (because classifier.predict returns float32)
        t0 = time.time()
        out_nodataVal = out_nodataVal if out_nodataVal is not None else image.nodata
        image_predicted = GeoArray(np.empty((image.rows, image.cols, classifier.tgt_n_bands), dtype=image.dtype),
                                   geotransform=image.gt, projection=image.prj, nodata=out_nodataVal,
                                   bandnames=['B%s' % i if len(i) == 2 else 'B0%s' % i for i in classifier.tgt_LBA])

        if classifier.n_clusters > 1 and self.classif_map.ndim > 2:
            dist_min, dist_max = self.distance_metrics.min(), self.distance_metrics.max()
            dist_norm = (self.distance_metrics - dist_min) / (dist_max - dist_min)
            weights = 1 - dist_norm
        else:
            weights = None

        # weights = None if self.classif_map.ndim == 2 else \
        #     1 - (self.distance_metrics / np.sum(self.distance_metrics, axis=2, keepdims=True))

        # if self.classif_map.ndim > 2:
        #     print(self.distance_metrics[0, 0, :])
        #     print(weights[0, 0, :])

        n_saturated_px = 0
        for ((rS, rE), (cS, cE)), im_tile in image.tiles(tilesize=(1000, 1000)):
            self.logger.info('Predicting tile ((%s, %s), (%s, %s))...' % (rS, rE, cS, cE))

            classif_map_tile = self.classif_map[rS: rE + 1, cS: cE + 1]  # integer array

            # predict!
            if self.classif_map.ndim == 2:
                im_tile_pred = \
                    classifier.predict(im_tile, classif_map_tile,
                                       nodataVal=out_nodataVal,
                                       cmap_nodataVal=cmap_nodataVal,
                                       cmap_unclassifiedVal=unclassified_pixVal)

            else:
                weights_tile = weights[rS: rE + 1, cS: cE + 1]  # float array

                im_tile_pred = \
                    classifier.predict_weighted_averages(im_tile, classif_map_tile, weights_tile,
                                                         nodataVal=out_nodataVal,
                                                         cmap_nodataVal=cmap_nodataVal,
                                                         cmap_unclassifiedVal=unclassified_pixVal)

            # set saturated pixels (exceeding the output data range with respect to the data type) to no-data
            if isinstance(image_predicted.dtype, np.integer):
                out_dTMin, out_dTMax = np.iinfo(image_predicted.dtype).min, np.iinfo(image_predicted.dtype).max
                if im_tile_pred.min() < out_dTMin or im_tile_pred.max() > out_dTMax:
                    mask_saturated = np.any(im_tile_pred > out_dTMax | im_tile_pred < out_dTMin, axis=2)
                    n_saturated_px += np.sum(mask_saturated)
                    im_tile_pred[mask_saturated] = out_nodataVal

            image_predicted[rS:rE + 1, cS:cE + 1] = im_tile_pred.astype(image_predicted.dtype)

        if n_saturated_px:
            self.logger.warning("%.2f %% of the predicted pixels are saturated and set to no-data."
                                % n_saturated_px / np.dot(*image_predicted.shape[:2]) * 100)

        # TODO add multiprocessing here? ML classifiers seem to use multiprocessing already
        # print(time.time() -t0)
        # t0 = time.time()
        # from multiprocessing import Pool
        # from geoarray.baseclasses import get_array_tilebounds
        # with Pool(self.CPUs, initializer=_mp_initializer, initargs=(image, self.classif_map, classifier)) as pool:
        #     tiles_pred = pool.starmap(_predict_tile_mp,
        #                               [(tilebounds, out_nodataVal, cmap_nodataVal)
        #                                for tilebounds in get_array_tilebounds(array_shape=image.shape,
        #                                                                       tile_shape=(1000, 1000))])
        #
        # for ((rS, rE), (cS, cE)), tile_pred in tiles_pred:
        #     image_predicted[rS: rE + 1, cS: cE + 1, :] = tile_pred
        #
        # print(time.time() - t0)

        self.logger.info('Total prediction time: %s' % time.strftime("%H:%M:%S", time.gmtime(time.time()-t0)))

        ###############################
        # complete prediction results #
        ###############################

        # re-apply nodata values to predicted result
        if image.nodata is not None:
            image_predicted[image.mask_nodata[:] == 0] = out_nodataVal

        # copy mask_nodata
        image_predicted.mask_nodata = image.mask_nodata

        # image_predicted.save(
        #     '/home/gfz-fe/scheffler/temp/SPECHOM_py/image_predicted_QRclust1_MinDist_noB9.bsq')
        # GeoArray(self.classif_map).save(
        #     '/home/gfz-fe/scheffler/temp/SPECHOM_py/classif_map_QRclust1_MinDist_noB9.bsq')

        # append some statistics regarding the homogenization
        cmap_vals, cmap_valcounts = np.unique(self.classif_map, return_counts=True)
        cmap_valfractions = cmap_valcounts / self.classif_map.size
        self.classif_map_fractions = dict(zip(list(cmap_vals), list(cmap_valfractions)))

        return image_predicted

    def compute_prediction_errors(self, im_predicted, cluster_classifier, nodataVal=None, cmap_nodataVal=None):
        # type: (Union[np.ndarray, GeoArray], Cluster_Learner, float, float) -> np.ndarray
        """Compute errors that quantify prediction inaccurracy per band and per pixel.

        :param im_predicted:        3D array representing the predicted image
        :param cluster_classifier:  instance of Cluster_Learner
        :param nodataVal:           no data value of the input image
                                    (auto-computed if not given or contained in im_predicted GeoArray)
                                    NOTE: The value is also used as output nodata value for the errors array.
        :param cmap_nodataVal:      no data value for the classification map
                                    in case more than one sub-classes are used for prediction
        :return:                    3D array (int16) representing prediction errors per band and pixel
        """
        im_predicted = im_predicted if isinstance(im_predicted, GeoArray) else GeoArray(im_predicted, nodata=nodataVal)
        im_predicted.nodata = nodataVal if nodataVal is not None else im_predicted.nodata  # might be auto-computed here

        for clf in cluster_classifier:
            if not len(clf.rmse_per_band) == GeoArray(im_predicted).bands:
                raise ValueError('The given classifier contains error statistics incompatible to the shape of the '
                                 'image.')
        if self.classif_map is None:
            raise RuntimeError('self.classif_map must be generated by running self.predict() beforehand.')

        if self.classif_map.ndim == 3:
            # FIXME: error computation does not work for kNN algorithms so far (self.classif_map is 3D instead of 2D)
            raise NotImplementedError('Error computation for 3-dimensional classification maps (e.g., due to kNN '
                                      'classification algorithms) is not yet implemented.')

        errors = np.empty_like(im_predicted)

        # iterate over all cluster labels and copy rmse values
        for pixVal in sorted(list(np.unique(self.classif_map))):
            if pixVal == cmap_nodataVal:
                continue

            self.logger.info('Inpainting error values for cluster #%s...' % pixVal)

            clf2use = cluster_classifier.MLdict[pixVal] if pixVal != -1 else cluster_classifier.global_clf
            rmse_per_band_int = np.round(clf2use.rmse_per_band, 0).astype(np.int16)
            errors[self.classif_map[:] == pixVal] = rmse_per_band_int

        # TODO validate this equation
        # errors = (errors * im_predicted[:] / 10000).astype(errors.dtype)

        # re-apply nodata values to predicted result
        if im_predicted.nodata is not None:
            # errors[im_predicted == im_predicted.nodata] = im_predicted.nodata
            errors[im_predicted.mask_nodata.astype(np.int8) == 0] = im_predicted.nodata

        # GeoArray(errors).save('/home/gfz-fe/scheffler/temp/SPECHOM_py/errors_LRclust1_MinDist_noB9_clusterpred.bsq')

        return errors

#
# _global_image, _global_classif_map, _global_classifier = None, None, None
#
#
# def _mp_initializer(image, classif_map, classifier):
#     global _global_image, _global_classif_map, _global_classifier
#     _global_image, _global_classif_map, _global_classifier = image, classif_map, classifier
#
#
# def _predict_tile_mp(tilebounds, out_nodataVal, cmap_nodataVal):
#     (rS, rE), (cS, cE) = tilebounds
#     im_tile = _global_image[rS: rE + 1, cS: cE + 1, :]
#     classif_map_tile = _global_classif_map[rS: rE + 1, cS: cE + 1]  # integer array
#     classifier = _global_classifier
#
#     # predict!
#     im_tile_pred = \
#         classifier.predict(im_tile, classif_map_tile,
#                            nodataVal=out_nodataVal, cmap_nodataVal=cmap_nodataVal).astype(_global_image.dtype)
#
#     return tilebounds, im_tile_pred



# -*- coding: utf-8 -*-

# spechomo, Spectral homogenization of multispectral satellite data
#
# Copyright (C) 2020  Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
#
# This software was developed within the context of the GeoMultiSens project funded
# by the German Federal Ministry of Education and Research
# (project grant code: 01 IS 14 010 A-C).
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version. Please note the following exception: `spechomo` depends on tqdm,
# which is distributed under the Mozilla Public Licence (MPL) v2.0 except for the
# files "tqdm/_tqdm.py", "setup.py", "README.rst", "MANIFEST.in" and ".gitignore".
# Details can be found here: https://github.com/tqdm/tqdm/blob/master/LICENCE.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import tempfile
import zipfile
from collections import OrderedDict
from pprint import pformat
from typing import Union, List, TYPE_CHECKING  # noqa F401  # flake8 issue
import json
import builtins

if TYPE_CHECKING:
    from matplotlib import pyplot as plt  # noqa F401  # flake8 issue

from tqdm import tqdm
import dill
import numpy as np
from pandas import DataFrame
from geoarray import GeoArray  # noqa F401  # flake8 issue

from .classifier_creation import get_filename_classifier_collection, get_machine_learner
from .exceptions import ClassifierNotAvailableError
from .utils import im2spectra, spectra2im


class Cluster_Learner(object):
    """
    A class that holds the machine learning classifiers for multiple spectral clusters as well as a global classifier.

    These classifiers can be applied to an input sensor image by using the predict method.
    """

    def __init__(self, dict_clust_MLinstances, global_classifier):
        # type: (Union[dict, ClassifierCollection], any) -> None
        """Get an instance of Cluster_Learner.

        :param dict_clust_MLinstances:  a dictionary of cluster specific machine learning classifiers
        :param global_classifier:       the global machine learning classifier to be applied at image positions with
                                        high spectral dissimilarity to the available cluster centers
        """
        self.cluster_pixVals = list(sorted(dict_clust_MLinstances.keys()))  # type: List[int]
        self.MLdict = OrderedDict((clust, dict_clust_MLinstances[clust]) for clust in self.cluster_pixVals)
        self.global_clf = global_classifier
        sample_MLinst = list(self.MLdict.values())[0]
        self.src_satellite = sample_MLinst.src_satellite
        self.src_sensor = sample_MLinst.src_sensor
        self.tgt_satellite = sample_MLinst.tgt_satellite
        self.tgt_sensor = sample_MLinst.tgt_sensor
        self.src_LBA = sample_MLinst.src_LBA
        self.tgt_LBA = sample_MLinst.tgt_LBA
        self.src_n_bands = sample_MLinst.src_n_bands
        self.tgt_n_bands = sample_MLinst.tgt_n_bands
        self.src_wavelengths = sample_MLinst.src_wavelengths
        self.tgt_wavelengths = sample_MLinst.tgt_wavelengths
        self.n_clusters = sample_MLinst.n_clusters
        self.cluster_centers = np.array([cc.cluster_center for cc in self.MLdict.values()])

    @classmethod
    def from_disk(cls, classifier_rootDir, method, n_clusters,
                  src_satellite, src_sensor, src_LBA, tgt_satellite, tgt_sensor, tgt_LBA, n_estimators=50):
        # type: (str, str, int, str, str, list, str, str, list, int) -> Cluster_Learner
        """Read a previously saved ClusterLearner from disk and return a ClusterLearner instance.

        Describe the classifier specifications with the given arguments.

        :param classifier_rootDir:  root directory of the classifiers
        :param method:              harmonization method
                                    'LR':   Linear Regression
                                    'RR':   Ridge Regression
                                    'QR':   Quadratic Regression
                                    'RFR':  Random Forest Regression  (50 trees; does not allow spectral sub-clustering)
        :param n_clusters:          number of clusters
        :param src_satellite:       source satellite, e.g., 'Landsat-8'
        :param src_sensor:          source sensor, e.g., 'OLI_TIRS'
        :param src_LBA:             source LayerBandsAssignment
        :param tgt_satellite:       target satellite, e.g., 'Landsat-8'
        :param tgt_sensor:          target sensor, e.g., 'OLI_TIRS'
        :param tgt_LBA:             target LayerBandsAssignment
        :param n_estimators:        number of estimators (only used in case of method=='RFR'
        :return:                    classifier instance loaded from disk
        """
        # get path of classifier zip archive
        args = (method, src_satellite, src_sensor, src_LBA, tgt_satellite, tgt_sensor, tgt_LBA)
        kw_clfinit = dict(n_estimators=n_estimators)

        if os.path.isfile(os.path.join(classifier_rootDir, '%s_classifiers.zip' % method)):
            # get cluster specific classifiers and store them in a ClassifierCollection dictionary
            dict_clust_MLinstances = cls._read_ClassifierCollection_from_zipFile(
                classifier_rootDir, *args, n_clusters=n_clusters, **kw_clfinit)

            # get the global classifier and add it as cluster label '-1'
            global_clf = cls._read_ClassifierCollection_from_zipFile(
                classifier_rootDir, *args, n_clusters=1, **kw_clfinit)[0]

        elif os.path.isdir(classifier_rootDir):
            # get cluster specific classifiers and store them in a ClassifierCollection dictionary
            dict_clust_MLinstances = cls._read_ClassifierCollection_from_unzippedFile(
                classifier_rootDir, *args, n_clusters=n_clusters, **kw_clfinit)

            # get the global classifier and add it as cluster label '-1'
            global_clf = cls._read_ClassifierCollection_from_unzippedFile(
                classifier_rootDir, *args, n_clusters=1, **kw_clfinit)[0]

        else:
            raise FileNotFoundError("No '%s' classifiers available at %s." % (method, classifier_rootDir))

        # create an instance of ClusterLearner based on the ClassifierCollection dictionary
        return Cluster_Learner(dict_clust_MLinstances, global_clf)

    @staticmethod
    def _read_ClassifierCollection_from_zipFile(classifier_rootDir, method, src_satellite, src_sensor,
                                                src_LBA, tgt_satellite, tgt_sensor, tgt_LBA, n_clusters,
                                                **kw_clfinit):
        # type: (str, str, str, str, list, str, str, list, int, dict) -> ClassifierCollection

        path_classifier_zip = os.path.join(classifier_rootDir, '%s_classifiers.zip' % method)

        # read requested classifier from zip archive and create a ClassifierCollection
        with zipfile.ZipFile(path_classifier_zip, "r") as zf, tempfile.TemporaryDirectory() as td:
            fName_clf = get_filename_classifier_collection(method, src_satellite, src_sensor, n_clusters=n_clusters,
                                                           **kw_clfinit)
            try:
                zf.extract(fName_clf, td)
            except KeyError:
                raise FileNotFoundError("No classifiers for %s %s with %d clusters contained in %s."
                                        % (src_satellite, src_sensor, n_clusters, path_classifier_zip))

            return Cluster_Learner._read_ClassifierCollection_from_unzippedFile(
                td, method, src_satellite, src_sensor, src_LBA,
                tgt_satellite, tgt_sensor, tgt_LBA, n_clusters)

    @staticmethod
    def _read_ClassifierCollection_from_unzippedFile(classifier_rootDir, method, src_satellite, src_sensor, src_LBA,
                                                     tgt_satellite, tgt_sensor, tgt_LBA, n_clusters, **kw_clfinit):
        # type: (str, str, str, str, list, str, str, list, int, dict) -> ClassifierCollection

        fName_clf_clustN = get_filename_classifier_collection(method, src_satellite, src_sensor,
                                                              n_clusters=n_clusters, **kw_clfinit)
        path_classifier = os.path.join(classifier_rootDir, fName_clf_clustN)

        # read requested classifier from zip archive and create a ClassifierCollection
        try:
            clf_collection = \
                ClassifierCollection(path_classifier)['__'.join(src_LBA)][tgt_satellite, tgt_sensor]['__'.join(tgt_LBA)]
        except KeyError:
            raise ClassifierNotAvailableError(method, src_satellite, src_sensor, src_LBA,
                                              tgt_satellite, tgt_sensor, tgt_LBA, n_clusters)

        # validation
        expected_MLtype = type(get_machine_learner(method))
        if len(clf_collection.keys()) != n_clusters:
            raise RuntimeError('Read classifier with %s clusters instead of %s.'
                               % (len(clf_collection.keys()), n_clusters))
        for label, ml in clf_collection.items():
            if not isinstance(ml, expected_MLtype):
                raise ValueError("The given dillFile %s contains a spectral cluster (label '%s') with a %s machine "
                                 "learner instead of the expected %s."
                                 % (os.path.basename(path_classifier), label, type(ml), expected_MLtype.__name__,))

        return clf_collection

    def __iter__(self):
        for cluster in self.cluster_pixVals:
            yield self.MLdict[cluster]

    def predict(self, im_src, cmap, nodataVal=None, cmap_nodataVal=None, cmap_unclassifiedVal=-1):
        # type: (Union[np.ndarray, GeoArray], np.ndarray, Union[int, float], Union[int, float], Union[int, float]) -> np.ndarray  # noqa
        """Predict target satellite spectral information using separate prediction coefficients for spectral clusters.

        :param im_src:          input image to be used for prediction
        :param cmap:            classification map that assigns each image spectrum to a corresponding cluster
                                -> must be a 2D np.ndarray with the same X-/Y-dimension like im_src
        :param nodataVal:       nodata value to be used to fill into the predicted image
        :param cmap_nodataVal:  nodata class value of the nodata class of the classification map
        :param cmap_unclassifiedVal:    'unclassified' class value of the nodata class of the classification map
        :return:
        """
        cluster_labels = sorted(list(np.unique(cmap)))

        im_pred = np.full((im_src.shape[0], im_src.shape[1], self.tgt_n_bands),
                          fill_value=nodataVal if nodataVal is not None else 0,
                          dtype=np.float32)

        if len(cluster_labels) > 1:
            # iterate over all cluster labels and apply corresponding machine learner parameters
            # to predict target spectra
            for pixVal in cluster_labels:
                if pixVal == cmap_nodataVal:
                    continue

                elif pixVal == cmap_unclassifiedVal:
                    # apply global homogenization coefficients
                    classifier = self.global_clf

                else:
                    # apply cluster specific homogenization coefficients
                    classifier = self.MLdict[pixVal]

                mask_pixVal = cmap == pixVal
                im_pred[mask_pixVal] = classifier.predict(im_src[mask_pixVal])

        else:
            # predict target spectra directly (much faster than the above algorithm)
            pixVal = cluster_labels[0]

            if pixVal == cmap_nodataVal:
                # im_src consists only of no data values
                pass  # im_pred keeps at nodataVal

            else:
                if pixVal == cmap_unclassifiedVal:
                    # apply global homogenization coefficients
                    classifier = self.global_clf
                else:
                    classifier = self.MLdict[pixVal]
                    assert classifier.clusterlabel == pixVal

                spectra = im2spectra(im_src)
                spectra_pred = classifier.predict(spectra)
                im_pred = spectra2im(spectra_pred, im_src.shape[0], im_src.shape[1])

        return im_pred  # float32 array

    def predict_weighted_averages(self, im_src, cmap_3D, weights_3D=None, nodataVal=None,
                                  cmap_nodataVal=None, cmap_unclassifiedVal=-1):
        # type: (Union[np.ndarray, GeoArray], np.ndarray, np.ndarray, Union[int, float], Union[int, float], Union[int, float]) -> np.ndarray  # noqa
        """Predict target satellite spectral information using separate prediction coefficients for spectral clusters.

        NOTE:   This version of the prediction function uses the prediction coefficients of multiple spectral clusters
                and computes the result as weighted average of them. Therefore, the classifcation map must assign
                multiple spectral cluster to each input pixel.

        # NOTE:   At unclassified pixels (cmap_3D[y,x,z>0] == -1) the prediction result using global coefficients
        #         is ignored in the weighted average. In that case the prediction result is based on the found valid
        #         spectral clusters and is not affected by the global coefficients (should improve prediction results).

        :param im_src:          input image to be used for prediction
        :param cmap_3D:         classification map that assigns each image spectrum to multiple corresponding clusters
                                -> must be a 3D np.ndarray with the same X-/Y-dimension like im_src
        :param weights_3D:
        :param nodataVal:       nodata value to be used to fill into the predicted image
        :param cmap_nodataVal:  nodata class value of the nodata class of the classification map
        :param cmap_unclassifiedVal:    'unclassified' class value of the nodata class of the classification map
        :return:
        """
        if not cmap_3D.ndim > 2:
            raise ValueError('Input classification map needs at least 2 bands to compute prediction results as'
                             'weighted averages.')

        if cmap_3D.shape != weights_3D.shape:
            raise ValueError("The input arrays 'cmap_3D' and 'weights_3D' need to have the same dimensions. "
                             "Received %s vs. %s." % (cmap_3D.shape, weights_3D.shape))

        # predict for each classification map band
        ims_pred_temp = []

        for band in range(cmap_3D.shape[2]):
            ims_pred_temp.append(
                self.predict(im_src, cmap_3D[:, :, band],
                             nodataVal=nodataVal,
                             cmap_nodataVal=cmap_nodataVal,
                             cmap_unclassifiedVal=cmap_unclassifiedVal
                             ))

        # merge classification results by weighted averaging
        nsamp, nbandpred, nbandscmap = np.dot(*weights_3D.shape[:2]), ims_pred_temp[0].shape[2], weights_3D.shape[2]
        weights = \
            np.ones((nsamp, nbandpred, nbandscmap)) if weights_3D is None else \
            np.tile(weights_3D.reshape(nsamp, 1, nbandscmap), (1, nbandpred, 1))  # nclust x n_tgt_bands x n_cmap_bands

        # set weighting of unclassified pixel positions to zero (except from the first cmap band)
        #   -> see NOTE #2 in the docstring
        # mask_unclassif = np.tile(cmap_3D.reshape(nsamp, 1, nbandscmap), (1, nbandpred, 1)) == cmap_unclassifiedVal
        # mask_unclassif[:, :, :1] = False  # if all other clusters are invalid, at least the first one is used for prediction # noqa
        # weights[mask_unclassif] = 0

        spectra_pred = np.average(np.dstack([im2spectra(im) for im in ims_pred_temp]), weights=weights, axis=2)
        im_pred = spectra2im(spectra_pred, tgt_rows=im_src.shape[0], tgt_cols=im_src.shape[1])

        return im_pred

    def plot_sample_spectra(self, cluster_label='all', include_mean_spectrum=True, include_median_spectrum=True,
                            ncols=5, **kw_fig):
        # type: (Union[str, int, List], bool, bool, int, dict) -> plt.figure
        from matplotlib import pyplot as plt  # noqa

        if isinstance(cluster_label, int):
            lbls2plot = [cluster_label]
        elif isinstance(cluster_label, list):
            lbls2plot = cluster_label
        elif cluster_label == 'all':
            lbls2plot = list(range(self.n_clusters))
        else:
            raise ValueError(cluster_label)

        # create a single plot
        if len(lbls2plot) == 1:
            if cluster_label == 'all':
                cluster_label = 0

            fig, axes = plt.figure(), None
            for i in range(100):
                plt.plot(self.src_wavelengths, self.MLdict[cluster_label].cluster_sample_spectra[i, :])

            plt.xlabel('wavelength [nm]')
            plt.ylabel('%s %s\nreflectance [0-10000]' % (self.src_satellite, self.src_sensor))
            plt.title('Cluster #%s' % cluster_label)
            plt.grid(lw=0.2)
            plt.ylim(0, 10000)

            if include_mean_spectrum:
                plt.plot(self.src_wavelengths, self.MLdict[cluster_label].cluster_center, c='black', lw=3)
            if include_median_spectrum:
                plt.plot(self.src_wavelengths, np.median(self.MLdict[cluster_label].cluster_sample_spectra, axis=0),
                         '--', c='black', lw=3)

        # create a plot with multiple subplots
        else:
            nplots = len(lbls2plot)
            ncols = nplots if nplots < ncols else ncols
            nrows = nplots // ncols if not nplots % ncols else nplots // ncols + 1
            figsize = (4 * ncols, 3 * nrows)
            fig, axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize, sharex='all', sharey='all',
                                     **kw_fig)

            for lbl, ax in tqdm(zip(lbls2plot, axes.flatten()), total=nplots):
                for i in range(100):
                    ax.plot(self.src_wavelengths, self.MLdict[lbl].cluster_sample_spectra[i, :], lw=1)

                if include_mean_spectrum:
                    ax.plot(self.src_wavelengths, self.MLdict[lbl].cluster_center, c='black', lw=2)
                if include_median_spectrum:
                    ax.plot(self.src_wavelengths, np.median(self.MLdict[lbl].cluster_sample_spectra, axis=0),
                            '--', c='black', lw=3)

                ax.grid(lw=0.2)
                ax.set_ylim(0, 10000)

                if ax.is_last_row():
                    ax.set_xlabel('wavelength [nm]')
                if ax.is_first_col():
                    ax.set_ylabel('%s %s\nreflectance [0-10000]' % (self.src_satellite, self.src_sensor))
                ax.set_title('Cluster #%s' % lbl)

        plt.tight_layout()
        plt.show()

        return fig, axes

    def _collect_stats(self, cluster_label):
        df = DataFrame(columns=['band', 'wavelength', 'RMSE', 'MAE', 'MAPE'])
        df.band = self.tgt_LBA
        df.wavelength = np.round(self.tgt_wavelengths, 1)
        df.RMSE = np.round(self.MLdict[cluster_label].rmse_per_band, 1)
        df.MAE = np.round(self.MLdict[cluster_label].mae_per_band, 1)
        df.MAPE = np.round(self.MLdict[cluster_label].mape_per_band, 1)

        overall_stats = dict(scores=self.MLdict[cluster_label].scores)

        return df, overall_stats

    def print_stats(self):
        from tabulate import tabulate

        for lbl in range(self.n_clusters):
            print('Cluster #%s:' % lbl)
            band_stats, overall_stats = self._collect_stats(lbl)
            print(overall_stats)
            print(tabulate(band_stats, headers=band_stats.columns))
            print()

    def to_jsonable_dict(self):
        """Create a dictionary containing a JSONable replicate of the current Cluster_Learner instance."""
        common_meta_keys = ['src_satellite', 'src_sensor', 'tgt_satellite', 'tgt_sensor', 'src_LBA', 'tgt_LBA',
                            'src_n_bands', 'tgt_n_bands', 'src_wavelengths', 'tgt_wavelengths', 'n_clusters']
        jsonable_dict = dict()
        decode_types_dict = dict()

        # get jsonable dict for global classifier and add decoding type hints
        jsonable_dict['classifier_global'] =\
            classifier_to_jsonable_dict(self.global_clf, skipkeys=common_meta_keys, include_typesdict=True)
        decode_types_dict['classifiers_all'] = jsonable_dict['classifier_global']['__decode_types']
        del jsonable_dict['classifier_global']['__decode_types']

        # get jsonable dicts for each classifier of self.MLdict and add corresponding decoding type hints
        jsonable_dict['classifiers_optimized'] =\
            {i: classifier_to_jsonable_dict(clf, skipkeys=common_meta_keys)
             for i, clf in self.MLdict.items()}

        # add common metadata and corresponding decoding type hints
        for k in common_meta_keys:
            jsonable_dict[k], decode_type = get_jsonable_value(getattr(self, k), return_typesdict=True)

            if decode_type:
                decode_types_dict[k] = decode_type

        jsonable_dict['__decode_types'] = decode_types_dict

        return jsonable_dict

    # def save_to_json(self, filepath):
    #     dict2save = dict(
    #         cluster_centers=self.cluster_centers.tolist(),
    #
    #     )
    #
    #     # Create json and save to file
    #     json_txt = json.dumps(dict2save, indent=4)
    #     with open(filepath, 'w') as file:
    #         file.write(json_txt)


class ClassifierCollection(object):
    def __init__(self, path_dillFile):
        with open(path_dillFile, 'rb') as inF:
            self.content = dill.load(inF)

    def __repr__(self):
        """Return the representation of ClassifierCollection.

        :return: e.g., "{'1__2__3__4__5__7': {('Landsat-5', 'TM'): {'1__2__3__4__5__7':
                        LinearRegression(copy_X=True, fit_intercept=True, n_jobs=1, normalize=False)}, ..."
        """
        return pformat(self.content)

    def __getitem__(self, item):
        """Get a specific item of the ClassifierCollection."""
        try:
            return self.content[item]
        except KeyError:
            raise(KeyError("The classifier has no key '%s'. Available keys are: %s"
                           % (item, repr(self))))

    # def save_to_json(self, filepath):
    #     a = 1
    #     pass


def get_jsonable_value(in_value, return_typesdict=False):
    if isinstance(in_value, np.ndarray):
        outval = in_value.tolist()
    elif isinstance(in_value, list):
        outval = np.array(in_value).tolist()
        # json.dumps(outval)
    else:
        outval = in_value

    # FIXME: In case of quadratic regression, there are some attributes that are not directly JSONable in this manner.

    # create a dictionary containing the data types needed for JSON decoding
    typesdict = dict()
    if return_typesdict and not isinstance(in_value, (str, int, float, bool)) and in_value is not None:
        typesdict['type'] = type(in_value).__name__

        if isinstance(in_value, np.ndarray):
            typesdict['dtype'] = in_value.dtype.name

        if isinstance(in_value, list):
            typesdict['dtype'] = type(in_value[0]).__name__

            if not len(set(type(vv).__name__ for vv in in_value)) == 1:
                raise RuntimeError('Lists containing different data types of list elements cannot be made '
                                   'jsonable without losses.')

    if return_typesdict:
        return outval, typesdict
    else:
        return outval


def classifier_to_jsonable_dict(clf, skipkeys: list = None, include_typesdict=False):
    from sklearn.linear_model import LinearRegression  # avoids static TLS error here

    if isinstance(clf, LinearRegression):
        jsonable_dict = dict(clftype='LR')
        typesdict = dict()

        for k, v in clf.__dict__.items():
            if skipkeys and k in skipkeys:
                continue

            if include_typesdict:
                jsonable_dict[k], typesdict[k] = get_jsonable_value(v, return_typesdict=True)
            else:
                jsonable_dict[k] = get_jsonable_value(v)

            # if valtype is np.ndarray:
            #     jsonable_dict[k] = dict(val=v.tolist(),
            #                        dtype=v.dtype.name)
            # elif valtype is list:
            #     jsonable_dict[k] = dict(val=np.array(v).tolist())
            # else:
            #     jsonable_dict[k] = dict(val=v)
            #
            # jsonable_dict[k]['valtype'] = valtype.__name__

    else:  # Ridge, Pipeline, RandomForestRegressor:
        # TODO
        raise NotImplementedError('At the moment, only LR classifiers can be serialized to JSON format.')

    if include_typesdict:
        jsonable_dict['__decode_types'] = {k: v for k, v in typesdict.items() if v}

    return jsonable_dict


def classifier_from_json_str(json_str):
    """Create a spectral harmonization classifier from a JSON string (JSON de-serialization).

    :param json_str:    the JSON string to be used for de-serialization
    :return:
    """
    from sklearn.linear_model import LinearRegression  # avoids static TLS error here

    in_dict = json.loads(json_str)

    if in_dict['clftype']['val'] == 'LR':
        clf = LinearRegression()
    else:
        raise NotImplementedError("Unknown object type '%s'." % in_dict['objecttype'])

    for k, v in in_dict.items():
        try:
            val2set = getattr(builtins, v['valtype'])(v['val'])
        except (AttributeError, KeyError):
            if v['valtype'] == 'ndarray':
                val2set = np.array(v['val']).astype(np.dtype(v['dtype']))
            else:
                raise TypeError("Unexpected object type '%s'." % v['valtype'])

        setattr(clf, k, val2set)

    return clf



# -*- coding: utf-8 -*-

# spechomo, Spectral homogenization of multispectral satellite data
#
# Copyright (C) 2020  Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
#
# This software was developed within the context of the GeoMultiSens project funded
# by the German Federal Ministry of Education and Research
# (project grant code: 01 IS 14 010 A-C).
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version. Please note the following exception: `spechomo` depends on tqdm,
# which is distributed under the Mozilla Public Licence (MPL) v2.0 except for the
# files "tqdm/_tqdm.py", "setup.py", "README.rst", "MANIFEST.in" and ".gitignore".
# Details can be found here: https://github.com/tqdm/tqdm/blob/master/LICENCE.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

from .version import __version__, __versionalias__   # noqa (E402 + F401)
from .prediction import SpectralHomogenizer  # noqa (E402 + F401)


__author__ = """Daniel Scheffler"""
__email__ = 'daniel.scheffler@gfz-potsdam.de'



# -*- coding: utf-8 -*-

# spechomo, Spectral homogenization of multispectral satellite data
#
# Copyright (C) 2020  Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
#
# This software was developed within the context of the GeoMultiSens project funded
# by the German Federal Ministry of Education and Research
# (project grant code: 01 IS 14 010 A-C).
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version. Please note the following exception: `spechomo` depends on tqdm,
# which is distributed under the Mozilla Public Licence (MPL) v2.0 except for the
# files "tqdm/_tqdm.py", "setup.py", "README.rst", "MANIFEST.in" and ".gitignore".
# Details can be found here: https://github.com/tqdm/tqdm/blob/master/LICENCE.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

from multiprocessing import cpu_count
from typing import Union, TYPE_CHECKING  # noqa F401  # flake8 issue
import os

if TYPE_CHECKING:
    from sklearn.cluster import KMeans

import dill
import numpy as np
from geoarray import GeoArray
from pandas import DataFrame
from specclassify import SAM_Classifier, classify_image

from .utils import im2spectra


class KMeansRSImage(object):
    """Class for clustering a given input image by using K-Means algorithm.

    NOTE: Based on the nodata value of the input GeoArray those pixels that have nodata values in some bands are
          ignored when computing the cluster coefficients. Nodata values would affect clustering result otherwise.
    """

    def __init__(self, im, n_clusters, sam_classassignment=False, CPUs=1, v=False):
        # type: (GeoArray, int, bool, Union[None, int], bool) -> None

        # privates
        self._clusters = None
        self._clustermap = None
        self._goodSpecMask = None
        self._spectra = None
        self._labels_with_nodata = None
        self._spectral_distances = None
        self._spectral_distances_with_nodata = None
        self._spectral_angles = None
        self._spectral_angles_with_nodata = None

        self.im = im
        self.n_clusters = n_clusters
        self.sam_classassignment = sam_classassignment
        self.CPUs = CPUs or cpu_count()
        self.v = v

    @classmethod
    def from_disk(cls, path_clf, im):
        # type: (str, GeoArray) -> KMeansRSImage
        """Get an instance of KMeansRSImage from a previously saved classifier.

        :param path_clf:    path of serialzed classifier (dill file)
        :param im:          path of the image cube belonging to that classifier
        :return: KMeansRSImage
        """
        with open(path_clf, 'rb') as inF:
            undilled = dill.load(inF)  # type: dict

        KM = KMeansRSImage(im,
                           undilled['clusters'].n_clusters,
                           undilled['clusters'].n_jobs,
                           undilled['clusters'].verbose)
        KM._clusters = undilled['clusters']
        KM._goodSpecMask = undilled['_goodSpecMask']
        KM._spectral_distances = undilled['_spectral_distances']

        return KM

    @property
    def goodSpecMask(self):
        if self._goodSpecMask is None:
            if self.im.nodata is not None:
                mask_nodata = im2spectra(self.im) == self.im.nodata
                goodSpecMask = np.all(~mask_nodata, axis=1)

                if True not in goodSpecMask:
                    raise RuntimeError('All spectra contain no data values in one or multiple bands and are therefore '
                                       'not usable for clustering. Clustering failed.')

                self._goodSpecMask = goodSpecMask
            else:
                self._goodSpecMask = np.ones((self.im.rows * self.im.cols), dtype=np.bool)

        return self._goodSpecMask

    @property
    def spectra(self):
        """Get spectra used for clustering (excluding spectra containing nodata values that would affect clustering)."""
        if self._spectra is None:
            self._spectra = im2spectra(self.im)[self.goodSpecMask, :]
        return self._spectra

    @property
    def n_spectra(self):
        """Get number of spectra used for clustering (excluding spectra containing nodata values)."""
        return int(np.sum(self.goodSpecMask))

    @property
    def clusters(self):
        # type: () -> KMeans
        if not self._clusters:
            self._clusters = self.compute_clusters()
        return self._clusters

    @clusters.setter
    def clusters(self, clusters):
        self._clusters = clusters

    @property
    def clustermap(self):
        if self._clustermap is None:
            self._clustermap = self.labels_with_nodata.reshape((self.im.rows, self.im.cols))

        return self._clustermap

    def compute_clusters(self, nmax_spectra=100000):
        """Compute the cluster means and labels.

        :param nmax_spectra:    maximum number of spectra to be included (pseudo-randomly selected (reproducable))
        :return:
        """
        from sklearn.cluster import KMeans  # avoids static TLS ImportError here
        from sklearn import __version__ as skver

        _old_environ = dict(os.environ)

        try:
            kwargs_kmeans = dict(n_clusters=self.n_clusters, random_state=0, verbose=self.v)
            # scikit-learn>0.23 uses all cores by default; number is adjustable via OMP_NUM_THREADS
            if float('%s.%s' % tuple(skver.split('.')[:2])) < 0.23:
                kwargs_kmeans['n_jobs'] = self.CPUs
            else:
                os.environ['OMP_NUM_THREADS '] = str(self.CPUs)

            # data reduction in case we have too many spectra
            if self.spectra.shape[0] > nmax_spectra:
                if self.v:
                    print('Reducing data...')
                idxs_specIncl = np.random.RandomState(seed=0).choice(range(self.n_spectra), nmax_spectra)
                idxs_specNotIncl = np.array(range(self.n_spectra))[~np.in1d(range(self.n_spectra), idxs_specIncl)]
                spectra_incl = self.spectra[idxs_specIncl, :]
                spectra_notIncl = self.spectra[idxs_specNotIncl, :]

                if self.v:
                    print('Fitting KMeans...')
                kmeans = KMeans(**kwargs_kmeans)
                distmatrix_incl = kmeans.fit_transform(spectra_incl)

                if self.v:
                    print('Computing full resolution labels...')
                labels = np.zeros((self.n_spectra,), dtype=kmeans.labels_.dtype)
                distances = np.zeros((self.n_spectra,), dtype=distmatrix_incl.dtype)

                labels[idxs_specIncl] = kmeans.labels_
                distances[idxs_specIncl] = np.min(distmatrix_incl, axis=1)

                if self.sam_classassignment:
                    # override cluster labels with labels computed via SAM (distances have be recomputed then)
                    print('Using SAM class assignment.')
                    SC = SAM_Classifier(kmeans.cluster_centers_, CPUs=self.CPUs)
                    im_sam_labels = SC.classify(self.im)
                    sam_labels = im_sam_labels.flatten()[self.goodSpecMask]
                    self._spectral_angles = SC.angles_deg.flatten()[self.goodSpecMask]

                    # update distances at those positions where SAM assigns different class labels
                    distsPos2update = labels != sam_labels
                    distsPos2update[idxs_specNotIncl] = True
                    distances[distsPos2update] = \
                        self.compute_euclidian_distance_for_labelled_spectra(
                            self.spectra[distsPos2update, :], sam_labels[distsPos2update], kmeans.cluster_centers_)

                    kmeans.labels_ = sam_labels

                else:
                    distmatrix_specNotIncl = kmeans.transform(spectra_notIncl)
                    labels[idxs_specNotIncl] = np.argmin(distmatrix_specNotIncl, axis=1)
                    distances[idxs_specNotIncl] = np.min(distmatrix_specNotIncl, axis=1)

                    kmeans.labels_ = labels

            else:
                if self.v:
                    print('Fitting KMeans...')
                kmeans = KMeans(**kwargs_kmeans)

                distmatrix = kmeans.fit_transform(self.spectra)
                kmeans_labels = kmeans.labels_
                distances = np.min(distmatrix, axis=1)

                if self.sam_classassignment:
                    # override cluster labels with labels computed via SAM (distances have be recomputed then)
                    print('Using SAM class assignment.')
                    SC = SAM_Classifier(kmeans.cluster_centers_, CPUs=self.CPUs)
                    im_sam_labels = SC.classify(self.im)
                    sam_labels = im_sam_labels.flatten()[self.goodSpecMask]
                    self._spectral_angles = SC.angles_deg.flatten()[self.goodSpecMask]

                    # update distances at those positions where SAM assigns different class labels
                    distsPos2update = kmeans_labels != sam_labels
                    distances[distsPos2update] = \
                        self.compute_euclidian_distance_for_labelled_spectra(
                            self.spectra[distsPos2update, :], sam_labels[distsPos2update], kmeans.cluster_centers_)
        finally:
            os.environ.clear()
            os.environ.update(_old_environ)

        self.clusters = kmeans
        self._spectral_distances = distances

        return self.clusters

    def apply_clusters(self, image):
        labels = self.clusters.predict(im2spectra(GeoArray(image)))
        return labels

    def compute_spectral_distances(self):
        self._spectral_distances = np.min(self.clusters.transform(self.spectra), axis=1)
        return self.spectral_distances

    @staticmethod
    def compute_euclidian_distance_2D(spectra, endmembers):
        n_samples, n_features = endmembers.shape

        if not spectra.shape[1] == endmembers.shape[1]:
            raise RuntimeError('Matrix dimensions are not aligned. Input spectra have %d bands but endmembers '
                               'have %d.' % (spectra.shape[1], endmembers.shape[1]))

        dists = np.zeros((spectra.shape[0], n_samples), np.float32)

        # loop over all endmember spectra and compute spectral angle for each input spectrum
        for n_sample in range(n_samples):
            train_spectrum = endmembers[n_sample, :].astype(np.float)
            diff = spectra - train_spectrum
            dists[:, n_sample] = np.sqrt((diff ** 2).sum(axis=1))

        return dists

    @staticmethod
    def compute_euclidian_distance_for_labelled_spectra(spectra, labels, endmembers):
        if not spectra.shape[1] == endmembers.shape[1]:
            raise RuntimeError('Matrix dimensions are not aligned. Input spectra have %d bands but endmembers '
                               'have %d.' % (spectra.shape[1], endmembers.shape[1]))

        dists = np.zeros((spectra.shape[0]), np.float32)

        # loop over all endmember spectra and compute spectral angle for each input spectrum
        for lbl in np.unique(labels):
            train_spectrum = endmembers[lbl, :].astype(np.float)
            mask_curlbl = labels == lbl
            spectra_curlbl = spectra[mask_curlbl, :]
            diff_curlbl = spectra_curlbl - train_spectrum
            dists[mask_curlbl] = np.sqrt((diff_curlbl ** 2).sum(axis=1))

        return dists

    def compute_spectral_angles(self):
        spectral_angles = classify_image(self.im, self.clusters.cluster_centers_, list(range(self.n_clusters)),
                                         'SAM', in_nodataVal=self.im.nodata, cmap_nodataVal=-9999,
                                         tiledims=(400, 400), CPUs=self.CPUs, return_distance=True,
                                         unclassified_pixVal=-1)[1]

        self._spectral_angles = spectral_angles.flatten()[self.goodSpecMask]
        return self.spectral_angles

    @property
    def labels(self):
        """Get labels for all clustered spectra (excluding spectra that contain nodata values)."""
        return self.clusters.labels_

    @property
    def labels_with_nodata(self):
        """Get the labels for all pixels (including those containing nodata values)."""
        if self._labels_with_nodata is None:
            if self.n_spectra == (self.im.rows * self.im.cols):
                self._labels_with_nodata = self.clusters.labels_
            else:
                labels = np.full_like(self.goodSpecMask, -9999, dtype=self.clusters.labels_.dtype)
                labels[self.goodSpecMask] = self.clusters.labels_
                self._labels_with_nodata = labels

        return self._labels_with_nodata

    @property
    def spectral_distances(self):
        """Get spectral distances for all pixels that don't contain nodata values."""
        if self._spectral_distances is None:
            self._spectral_distances = self.compute_spectral_distances()

        return self._spectral_distances

    @property
    def spectral_distances_with_nodata(self):
        if self._clusters is not None and self._spectral_distances_with_nodata is None:
            if self.n_spectra == (self.im.rows * self.im.cols):
                self._spectral_distances_with_nodata = self.spectral_distances
            else:
                dists = np.full_like(self.goodSpecMask, np.nan, dtype=np.float)
                dists[self.goodSpecMask] = self.spectral_distances
                self._spectral_distances_with_nodata = dists

        return self._spectral_distances_with_nodata

    @property
    def spectral_angles(self):
        """Get spectral angles in degrees for all pixels that don't contain nodata values."""
        if self._spectral_angles is None:
            self._spectral_angles = self.compute_spectral_angles()

        return self._spectral_angles

    @property
    def spectral_angles_with_nodata(self):
        if self._clusters is not None and self._spectral_angles_with_nodata is None:
            if self.n_spectra == (self.im.rows * self.im.cols):
                self._spectral_angles_with_nodata = self.spectral_angles
            else:
                angles = np.full_like(self.goodSpecMask, np.nan, dtype=np.float)
                angles[self.goodSpecMask] = self.spectral_angles
                self._spectral_angles_with_nodata = angles

        return self._spectral_angles_with_nodata

    def dump(self, path_out):
        with open(path_out, 'wb') as outF:
            dill.dump(dict(
                clusters=self.clusters,
                _goodSpecMask=self._goodSpecMask,
                _spectral_distances=self._spectral_distances),
                outF)

    def save_clustermap(self, path_out, **kw_save):
        GeoArray(self.clustermap, geotransform=self.im.gt, projection=self.im.prj, nodata=-9999)\
            .save(path_out, **kw_save)

    def plot_cluster_centers(self, figsize=(15, 5)):
        # type: (tuple) -> None
        """Show a plot of the cluster center signatures.

        :param figsize:     figure size (inches)
        """
        from matplotlib import pyplot as plt

        plt.figure(figsize=figsize)
        for i, center_signature in enumerate(self.clusters.cluster_centers_):
            plt.plot(range(1, self.im.bands + 1), center_signature, label='Cluster #%s' % (i + 1))

        plt.title('KMeans cluster centers for %s clusters' % self.n_clusters)
        plt.xlabel('Spectral band')
        plt.ylabel('Pixel values')
        plt.legend(loc='upper right')

        plt.show()

    def plot_cluster_histogram(self, figsize=(15, 5)):
        # type: (tuple) -> None
        """Show a histogram indicating the proportion of each cluster label in percent.

        :param figsize:     figure size (inches)
        """
        from matplotlib import pyplot as plt

        # grab the number of different clusters and create a histogram
        # based on the number of pixels assigned to each cluster
        numLabels = np.arange(0, len(np.unique(self.clusters.labels_)) + 1)
        hist, bins = np.histogram(self.clusters.labels_, bins=numLabels)

        # normalize the histogram, such that it sums to 100
        hist = hist.astype("float")
        hist /= hist.sum() / 100

        # plot the histogram as bar plot
        plt.figure(figsize=figsize)

        plt.bar(bins[:-1], hist, width=1)

        plt.title('Proportion of cluster labels (%s clusters)' % self.n_clusters)
        plt.xlabel('# Cluster')
        plt.ylabel('Proportion [%]')

        plt.show()

    def plot_clustermap(self, figsize=None):
        # type: (tuple) -> None
        """Show a the clustered image.

        :param figsize:     figure size (inches)
        """
        from matplotlib import pyplot as plt

        plt.figure(figsize=figsize)
        rows, cols = self.clustermap.shape[:2]

        image2plot = self.clustermap if self.im.nodata is None else np.ma.masked_equal(self.clustermap, self.im.nodata)

        plt.imshow(image2plot, plt.get_cmap('prism'), interpolation='none', extent=(0, cols, rows, 0))
        plt.show()

    def get_random_spectra_from_each_cluster(self, samplesize=50, max_distance=None, max_angle=None,
                                             nmin_unique_spectra=50):
        # type: (int, Union[int, float, str], Union[int, float, str], int) -> dict
        """Return a given number of spectra randomly selected within each cluster.

        E.g., 50 spectra belonging to cluster 1, 50 spectra belonging to cluster 2 and so on.

        :param samplesize:  number of spectra to be randomly selected from each cluster
        :param max_distance:    spectra with a larger spectral distance than the given value will be excluded from
                                random sampling.
                                - if given as string like '20%', the maximum spectral distance is computed as 20%
                                percentile within each cluster
        :param max_angle:       spectra with a larger spectral angle than the given value will be excluded from
                                random sampling.
                                - if given as string like '20%', the maximum spectral angle is computed as 20%
                                percentile within each cluster
        :param nmin_unique_spectra:   in case a cluster has less than the given number, do not use its spectra
                                      (return missing values)
        :return:
        """
        # get DataFrame with columns [cluster_label, B1, B2, B3, ...]
        df = DataFrame(self.spectra, columns=['B%s' % band for band in range(1, self.im.bands + 1)], )
        df.insert(0, 'cluster_label', self.clusters.labels_)

        if max_angle is not None:
            if not (isinstance(max_angle, (int, float)) and max_angle > 0) and \
               not (isinstance(max_angle, str) and max_angle.endswith('%')):
                raise ValueError(max_angle)
            if isinstance(max_angle, str):
                max_angle = np.percentile(self.spectral_angles, float(max_angle.split('%')[0].strip()))
            df.insert(1, 'spectral_angle', self.spectral_angles)

        if max_distance is not None:
            if not (isinstance(max_distance, (int, float)) and 0 < max_distance < 100) and \
               not (isinstance(max_distance, str) and max_distance.endswith('%')):
                raise ValueError(max_distance)
            if isinstance(max_distance, str):
                max_distance = np.percentile(self.spectral_distances, float(max_distance.split('%')[0].strip()))
            df.insert(1, 'spectral_distance', self.spectral_distances)

        # get random sample from each cluster and generate a dict like {cluster_label: random_sample}
        # NOTE: nodata label is skipped
        random_samples = dict()
        for label in range(self.n_clusters):
            if max_angle is None and max_distance is None:
                cluster_subset = df[df.cluster_label == label].loc[:, 'B1':]
            else:
                cluster_subset = df[df.cluster_label == label]

                # if self.n_clusters > 1:
                # filter by spectral angle
                if len(cluster_subset.index) >= nmin_unique_spectra and max_angle is not None:
                    cluster_subset = cluster_subset[cluster_subset.spectral_angle < max_angle]

                # filter by spectral distance
                if len(cluster_subset.index) >= nmin_unique_spectra and max_distance is not None:
                    cluster_subset = cluster_subset[cluster_subset.spectral_distance < max_distance]

                cluster_subset = cluster_subset.loc[:, 'B1':]

                # handle clusters with less than nmin_unique_spectra in there
                if len(cluster_subset.index) < nmin_unique_spectra:
                    if len(cluster_subset.index) > 0:
                        # don't use the cluster (return nodata)
                        cluster_subset[:] = -9999

                    else:
                        # cluster_subset is empty after filtering -> return nodata
                        cluster_subset.loc[0] = [-9999] * len(cluster_subset.columns)

            # get random sample while filling it with duplicates of the same sample when cluster has not enough spectra
            random_samples[label] = np.array(cluster_subset.sample(samplesize, replace=True, random_state=20))

        return random_samples

    def get_purest_spectra_from_each_cluster(self, samplesize=50):
        # type: (int) -> dict
        """Return a given number of spectra directly surrounding the center of each cluster.

        E.g., 50 spectra belonging to cluster 1, 50 spectra belonging to cluster 2 and so on.

        :param samplesize:  number of spectra to be selected from each cluster
        :return:
        """
        # get DataFrame with columns [cluster_label, B1, B2, B3, ...]
        df = DataFrame(self.spectra, columns=['B%s' % band for band in range(1, self.im.bands + 1)], )
        df.insert(0, 'cluster_label', self.clusters.labels_)
        df.insert(1, 'spectral_distance', self.spectral_distances)

        # get random sample from each cluster and generate a dict like {cluster_label: random_sample}
        random_samples = dict()
        for label in range(self.n_clusters):
            cluster_subset = df[df.cluster_label == label].loc[:, 'spectral_distance':]

            # catch the case that the cluster does not contain enough spectra (duplicate existing ones)
            if len(cluster_subset.index) < samplesize:
                cluster_subset = cluster_subset.sample(samplesize, replace=True, random_state=20)

            cluster_subset_sorted = cluster_subset.sort_values(by=['spectral_distance'], ascending=True)
            random_samples[label] = np.array(cluster_subset_sorted.loc[:, 'B1':][:samplesize])

        return random_samples



# -*- coding: utf-8 -*-

# spechomo, Spectral homogenization of multispectral satellite data
#
# Copyright (C) 2020  Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
#
# This software was developed within the context of the GeoMultiSens project funded
# by the German Federal Ministry of Education and Research
# (project grant code: 01 IS 14 010 A-C).
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version. Please note the following exception: `spechomo` depends on tqdm,
# which is distributed under the Mozilla Public Licence (MPL) v2.0 except for the
# files "tqdm/_tqdm.py", "setup.py", "README.rst", "MANIFEST.in" and ".gitignore".
# Details can be found here: https://github.com/tqdm/tqdm/blob/master/LICENCE.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

from multiprocessing import Pool
from typing import Union, TYPE_CHECKING  # noqa F401  # flake8 issue

import numpy as np
from geoarray import GeoArray
from geoarray.baseclasses import get_array_tilebounds

from .logging import SpecHomo_Logger

if TYPE_CHECKING:
    from pyrsr import RSR  # noqa F401  # flake8 issue


class SpectralResampler(object):
    """Class for spectral resampling of a single spectral signature (1D-array) or an image (3D-array)."""

    def __init__(self, wvl_src, rsr_tgt, logger=None):
        # type: (np.ndarray, RSR, str) -> None
        """Get an instance of the SpectralResampler class.

        :param wvl_src:     center wavelength positions of the source spectrum
        :param rsr_tgt:     relative spectral response of the target instrument as an instance of
                            pyrsr.RelativeSpectralRespnse.
        """
        # privates
        self._wvl_1nm = None
        self._rsr_1nm = {}

        wvl_src = np.array(wvl_src, dtype=np.float).flatten()
        if rsr_tgt.wvl_unit != 'nanometers':
            rsr_tgt.convert_wvl_unit()

        self.wvl_src_nm = wvl_src if max(wvl_src) > 100 else wvl_src * 1000
        self.rsr_tgt = rsr_tgt
        self.logger = logger or SpecHomo_Logger(__name__)  # must be pickable

    @property
    def wvl_1nm(self):
        # spectral resampling of input image to 1 nm resolution
        if self._wvl_1nm is None:
            self._wvl_1nm = np.arange(np.ceil(self.wvl_src_nm.min()), np.floor(self.wvl_src_nm.max()), 1)
        return self._wvl_1nm

    @property
    def rsr_1nm(self):
        if not self._rsr_1nm:
            from scipy.interpolate import interp1d  # import here to avoid static TLS ImportError

            for band in self.rsr_tgt.bands:
                # resample RSR to 1 nm
                self._rsr_1nm[band] = \
                    interp1d(self.rsr_tgt.rsrs_wvl, self.rsr_tgt.rsrs[band],
                             bounds_error=False, fill_value=0, kind='linear')(self.wvl_1nm)

                # validate
                assert len(self._rsr_1nm[band]) == len(self.wvl_1nm)

        return self._rsr_1nm

    def resample_signature(self, spectrum, scale_factor=10000, nodataVal=None, alg_nodata='radical', v=False):
        # type: (np.ndarray, int, Union[int, float], str, bool) -> np.ndarray
        """Resample the given spectrum according to the spectral response functions of the target instument.

        :param spectrum:        spectral signature data
        :param scale_factor:    the scale factor to apply to the given spectrum when it is plotted (default: 10000)
        :param nodataVal:       no data value to be respected during resampling
        :param alg_nodata:      algorithm how to deal with pixels where the spectral bands of the source image
                                contain nodata within the spectral response of a target band
                                'radical':      set output band to nodata
                                'conservative': use existing spectral information and ignore nodata
                                                (might alter the outpur spectral information,
                                                 e.g., at spectral absorption bands)
        :param v:               enable verbose mode (shows a plot of the resampled spectrum) (default: False)
        :return:    resampled spectral signature
        """
        if not spectrum.ndim == 1:
            raise ValueError("The array of the given spectral signature must be 1-dimensional. "
                             "Received a %s-dimensional array." % spectrum.ndim)
        spectrum = np.array(spectrum, dtype=np.float).flatten()
        assert spectrum.size == self.wvl_src_nm.size

        # convert spectrum to one multispectral image pixel and resample it
        spectrum_rsp = \
            self.resample_image(spectrum.reshape(1, 1, spectrum.size),
                                tiledims=(1, 1),
                                nodataVal=nodataVal,
                                alg_nodata=alg_nodata,
                                CPUs=1)\
            .ravel()

        if v:
            from matplotlib import pyplot as plt

            plt.figure()
            for band in self.rsr_tgt.bands:
                plt.plot(self.wvl_1nm, self.rsr_1nm[band] / max(self.rsr_1nm[band]))
            plt.plot(self.wvl_src_nm, spectrum / scale_factor, '.')
            plt.plot(self.rsr_tgt.wvl, spectrum_rsp / scale_factor, 'x', color='r')
            plt.show()

        return spectrum_rsp

    def resample_spectra(self, spectra, chunksize=200, nodataVal=None, alg_nodata='radical', CPUs=None):
        # type: (Union[GeoArray, np.ndarray], int, Union[int, float], str, Union[None, int]) -> np.ndarray
        """Resample the given spectral signatures according to the spectral response functions of the target instrument.

        :param spectra:     spectral signatures, provided as 2D array
                            (rows: spectral samples, columns: spectral information / bands)
        :param chunksize:   defines how many spectral signatures are resampled per CPU
        :param nodataVal:   no data value to be respected during resampling
        :param alg_nodata:  algorithm how to deal with pixels where the spectral bands of the source image
                            contain nodata within the spectral response of a target band
                            'radical':      set output band to nodata
                            'conservative': use existing spectral information and ignore nodata
                                            (might alter the outpur spectral information,
                                             e.g., at spectral absorption bands)
        :param CPUs:        CPUs to use for processing
        """
        # input validation
        if not spectra.ndim == 2:
            ValueError("The given spectra array must be 2-dimensional. Received a %s-dimensional array."
                       % spectra.ndim)
        assert spectra.shape[1] == self.wvl_src_nm.size

        # convert spectra to one multispectral image column
        im_col = spectra.reshape(spectra.shape[0], 1, spectra.shape[1])

        im_col_rsp = self.resample_image(im_col,
                                         tiledims=(1, chunksize),
                                         nodataVal=nodataVal,
                                         alg_nodata=alg_nodata,
                                         CPUs=CPUs)
        spectra_rsp = im_col_rsp.reshape(im_col_rsp.shape[0], im_col_rsp.shape[2])

        return spectra_rsp

    def resample_image(self, image_cube, tiledims=(20, 20), nodataVal=None, alg_nodata='radical', CPUs=None):
        # type: (Union[GeoArray, np.ndarray], tuple, Union[int, float], str, Union[None, int]) -> np.ndarray
        """Resample the given spectral image cube according to the spectral response functions of the target instrument.

        :param image_cube:      image (3D array) containing the spectral information in the third dimension
        :param tiledims:        dimension of tiles to be used during computation (rows, columns)
        :param nodataVal:       nodata value of the input image
        :param alg_nodata:      algorithm how to deal with pixels where the spectral bands of the source image
                                contain nodata within the spectral response of a target band
                                'radical':      set output band to nodata
                                'conservative': use existing spectral information and ignore nodata
                                                (might alter the output spectral information,
                                                 e.g., at spectral absorption bands)
        :param CPUs:            CPUs to use for processing
        :return:    resampled spectral image cube
        """
        # input validation
        if not image_cube.ndim == 3:
            raise ValueError("The given image cube must be 3-dimensional. Received a %s-dimensional array."
                             % image_cube.ndim)
        assert image_cube.shape[2] == self.wvl_src_nm.size

        image_cube = GeoArray(image_cube)

        (R, C), B = image_cube.shape[:2], len(self.rsr_tgt.bands)
        image_rsp = np.zeros((R, C, B), dtype=image_cube.dtype)

        initargs = image_cube, self.rsr_tgt, self.rsr_1nm, self.wvl_src_nm, self.wvl_1nm
        if CPUs is None or CPUs > 1:
            with Pool(CPUs, initializer=_initializer_mp, initargs=initargs) as pool:
                tiles_rsp = pool.starmap(_resample_tile_mp,
                                         [(bounds, nodataVal, alg_nodata)
                                          for bounds in get_array_tilebounds(image_cube.shape, tiledims)])

        else:
            _initializer_mp(*initargs)
            tiles_rsp = [_resample_tile_mp(bounds, nodataVal, alg_nodata)
                         for bounds in get_array_tilebounds(image_cube.shape, tiledims)]

        for ((rS, rE), (cS, cE)), tile_rsp in tiles_rsp:
            image_rsp[rS: rE + 1, cS: cE + 1, :] = tile_rsp

        return image_rsp


_globs_mp = dict()


def _initializer_mp(image_cube, rsr_tgt, rsr_1nm, wvl_src_nm, wvl_1nm):
    # type: (np.ndarray, RSR, RSR, np.ndarray, np.ndarray) -> None
    global _globs_mp
    _globs_mp.update(dict(
        image_cube=image_cube,
        rsr_tgt=rsr_tgt,
        rsr_1nm=rsr_1nm,
        wvl_src_nm=wvl_src_nm,
        wvl_1nm=wvl_1nm,
    ))


def _resample_tile_mp(tilebounds, nodataVal=None, alg_nodata='radical'):
    # TODO speed up by using numba as described here https://krstn.eu/fast-linear-1D-interpolation-with-numba/

    from scipy.interpolate import interp1d

    (rS, rE), (cS, cE) = tilebounds

    # get global share variables
    tiledata = _globs_mp['image_cube'][rS: rE + 1, cS: cE + 1, :]  # type: Union[GeoArray, np.ndarray]
    rsr_tgt = _globs_mp['rsr_tgt']
    rsr_1nm = _globs_mp['rsr_1nm']
    wvl_src_nm = _globs_mp['wvl_src_nm']
    wvl_1nm = _globs_mp['wvl_1nm']

    if alg_nodata not in ['radical', 'conservative']:
        raise ValueError(alg_nodata)

    tile_rsp = np.zeros((*tiledata.shape[:2], len(rsr_tgt.bands)), dtype=tiledata.dtype)

    if nodataVal is not None:
        if np.isfinite(nodataVal):
            _mask_bool3d = tiledata == nodataVal
            mask_anynodata = np.any(_mask_bool3d, axis=2)
            mask_allnodata = np.all(_mask_bool3d, axis=2)
        else:
            raise NotImplementedError(nodataVal)

        tiledata = tiledata.astype(np.float)
        tiledata[_mask_bool3d] = np.nan

        # fill pixels with all bands nodata
        tile_rsp[mask_allnodata] = nodataVal

        # upsample spectra containing data to 1nm and get nan-mask
        tilespectra_1nm = interp1d(wvl_src_nm, tiledata[~mask_allnodata],
                                   axis=1, bounds_error=False, fill_value=0, kind='linear')(wvl_1nm)
        isnan = np.isnan(tilespectra_1nm)
        nan_in_spec = np.any(isnan, axis=1)

        # compute resampled values for pixels without nodata
        tilespectra_1nm_nonan = tilespectra_1nm[~nan_in_spec, :]
        for band_idx, band in enumerate(rsr_tgt.bands):
            # compute the resampled image cube (np.average computes the weighted mean value)
            if rsr_1nm[band].sum() != 0:
                res = np.average(tilespectra_1nm_nonan, weights=rsr_1nm[band], axis=1)
                # NOTE: rounding here is important to prevent -9999. converted to -9998
                if not np.issubdtype(tile_rsp.dtype, np.floating):
                    res = np.around(res)
            else:
                res = nodataVal

            tile_rsp[~mask_anynodata, band_idx] = res

        # compute resampled values for pixels with nodata
        tilespectra_1nm_withnan = tilespectra_1nm[nan_in_spec, :]
        tilespectra_1nm_withnan_ma = np.ma.masked_invalid(tilespectra_1nm_withnan)
        mask_withnan = mask_anynodata & ~mask_allnodata
        isnan_sub = isnan[nan_in_spec, :]

        for band_idx, band in enumerate(rsr_tgt.bands):
            res_ma = np.ma.average(tilespectra_1nm_withnan_ma, axis=1, weights=rsr_1nm[band])

            # in case all 1nm bands for the current band are NaN, the resulting average will also be NaN => fill
            res = np.ma.filled(res_ma, nodataVal)

            if alg_nodata == 'radical':
                # set those output values to nodata where the input bands within the target RSR contain any nodata
                badspec = np.any(isnan_sub & (rsr_1nm[band] > 0), axis=1)
                res[badspec] = nodataVal

            if not np.issubdtype(tile_rsp.dtype, np.floating):
                res = np.around(res)

            tile_rsp[mask_withnan, band_idx] = res

    else:
        # spectral resampling of input image to 1 nm resolution
        tile_1nm = interp1d(wvl_src_nm, tiledata,
                            axis=2, bounds_error=False, fill_value=0, kind='linear')(wvl_1nm)

        for band_idx, band in enumerate(rsr_tgt.bands):
            # compute the resampled image cube (np.average computes the weighted mean value)
            res = np.average(tile_1nm, weights=rsr_1nm[band], axis=2)
            # NOTE: rounding here is important to prevent -9999. converted to -9998
            tile_rsp[:, :, band_idx] = res if np.issubdtype(tile_rsp.dtype, np.floating) else np.around(res)

    return tilebounds, tile_rsp
